(* Declarations and code for the Virtual Machine (MyVM) *)

signature VM_ARCH = 
sig
        type machinetype
        type mtypes
        type instruction
        exception compiletime_error
        exception runtime_error
        val statictypeof : exp -> mtypes
        val compile : exp -> instruction list
        val stack : machinetype list ref
        val table : (string * machinetype) list ref
        val init : unit -> unit
        val exec : instruction -> unit
        val execute : instruction list -> unit
        val st : unit -> machinetype list
end;

structure MyVM:VM_ARCH =
struct
        (* Datatype for representing the values of different types supported by the 
         * hardware of the machine. 
         * m2s : function for displaying these types.
         *)
        datatype machinetype = MI of int
                             | MF of real
                             | MB of bool
                             | MS of string
                             ;
        fun m2s (MI i) = "MI "^(Int.toString i)
           |m2s (MF r) = "MF "^(Real.toString r)
           |m2s (MB b) = "MB "^(Bool.toString b)
           |m2s (MS s) = "MS "^(s)
           ;

        (* Datatype for representing the types supported by the 
         * hardware of the machine.
         *)
        datatype mtypes = MINT | MFLOAT | MBOOL | MSTR | MDUMMY;

        (* The Instruction Set Architecture of the machine *)
        datatype instruction = LDI of int
                             | LDR of real
                             | LDB of bool
                             | LDS of string
                             | LDV of string
                             | VM_ADD | VM_SUB | VM_MUL | VM_DIV | VM_MOD
                             | VM_AND | VM_OR | VM_NOT | VM_IMP | VM_EQI
                             | VM_EQ | VM_LT | VM_LTE
                             | VAR of string
                             | END
                             ;
        (* Exceptions. 2 are compile-time. 3 are runtime. *)
        exception tycon_mismatch;                            
        exception compiletime_error;
        exception stack_empty;
        exception undefined_variable of string;
        exception math_error;
        exception runtime_error;

        (* The Physical architecture. *)
        (* A stack on which operations are performed. *)
        val stack = ref ([]:machinetype list);
        (* A table that keeps track of the variables. *)
        val table = ref ([]:(string * machinetype) list);
        
        (* Initialises the VM *)
        fun init () = (
                        stack := [];
                        table := [("c",MF 299792458.0),("Pi",MF Math.pi),("e",MF Math.e)]
                      );

        (* For value assignment to variables *)
        local
        fun asgn s n [] = [(s,n)]
           |asgn s n ((s1,n1)::l) = if(s = s1) then ((s1,n)::l) else ((s1,n1)::(asgn s n l))
        in
        fun assign s n = (table := (asgn s n (!table)); n)
        end;
        
        (* For table look up *)
        local 
        fun search s [] = raise (undefined_variable s)
           |search s ((s1,n)::l) = if(s = s1) then n else (search s l)
        in
        fun lookup v = (search v (!table))
        end;

        (* The Static type checker. Variables are assigned DUMMY type. Resolves
         * any type inconsistancy during compile time. 
         *)
        local
        fun jointype _ MDUMMY = MDUMMY
           |jointype MDUMMY _ = MDUMMY
           |jointype t1 t2 = if (t1=t2) then t1 else raise tycon_mismatch
        in
        fun statictypeof (N n) = MINT
           |statictypeof (R r) = MFLOAT
           |statictypeof (B b) = MBOOL
           |statictypeof (S s) = MSTR
           |statictypeof (V v) = MDUMMY
           |statictypeof (PLUS(e1,e2)) = let 
                                                val t = (jointype (statictypeof e1) (statictypeof e2))
                                         in
                                                if(t=MBOOL) then raise tycon_mismatch else t
                                         end
           |statictypeof (MINUS(e1,e2)) = statictypeof(PLUS(e1,e2))
           |statictypeof (MULT(e1,e2)) = statictypeof(PLUS(e1,e2))
           |statictypeof (DIV(e1,e2)) = statictypeof(PLUS(e1,e2))
           |statictypeof (MOD(e1,e2)) =  let 
                                                val t = (jointype (statictypeof e1) (statictypeof e2))
                                         in
                                                if(t=MBOOL orelse t=MFLOAT orelse t=MSTR) then raise tycon_mismatch else t
                                         end
           |statictypeof (AND(e1,e2)) =  let 
                                                val t = (jointype (statictypeof e1) (statictypeof e2))
                                         in
                                                if(t=MINT orelse t=MFLOAT orelse t=MSTR) then raise tycon_mismatch else t
                                         end
           |statictypeof (OR(e1,e2)) = statictypeof(AND(e1,e2))
           |statictypeof (NOT e) = let 
                                          val t = (statictypeof e)
                                   in
                                          if(t=MINT orelse t=MFLOAT) then raise tycon_mismatch else t
                                   end
           |statictypeof (IMP(e1,e2)) = statictypeof(AND(e1,e2))
           |statictypeof (EQI(e1,e2)) = statictypeof(AND(e1,e2))
           |statictypeof (EQ(e1,e2)) = let
                                                val t = (jointype (statictypeof e1) (statictypeof e2))
                                       in
                                                if(t<>MBOOL) then MBOOL else raise tycon_mismatch
                                       end
           |statictypeof (LT(e1,e2)) = statictypeof(EQ(e1,e2))
           |statictypeof (LTE(e1,e2)) = statictypeof(EQ(e1,e2))
           |statictypeof (GT(e1,e2)) = statictypeof(EQ(e1,e2))
           |statictypeof (GTE(e1,e2)) = statictypeof(EQ(e1,e2))
           |statictypeof (ASMT(s,e)) = (statictypeof e)
           |statictypeof (STMT e) = (statictypeof e)
        end;

        (* The Compiler. Generates Target machine code for MyVM from the AST. 
         * Compiled out is a sequence of instructions.
         *)
        local
        val symt = ref [""]
        fun insert s [] = [s]
           |insert s (x::xs) = if (x=s) then (x::xs) else (x::(insert s xs))
        fun search s [] = false
           |search s (x::xs) = (x=s) orelse (search s xs)

        fun compiler (N n) = [LDI(n)]
           |compiler (R r) = [LDR(r)]
           |compiler (B b) = [LDB(b)]
           |compiler (S s) = [LDS(s)]
           |compiler (V v) = if(search v (!symt)) then [LDV(v)] 
                             else raise (undefined_variable v)
           |compiler (PLUS(e1,e2)) = (compiler e1)@(compiler e2)@[VM_ADD]
           |compiler (MINUS(e1,e2)) = (compiler e1)@(compiler e2)@[VM_SUB]
           |compiler (MULT(e1,e2)) = (compiler e1)@(compiler e2)@[VM_MUL]
           |compiler (DIV(e1,e2)) = (compiler e1)@(compiler e2)@[VM_DIV]
           |compiler (MOD(e1,e2)) = (compiler e1)@(compiler e2)@[VM_MOD]
           |compiler (AND(e1,e2)) = (compiler e1)@(compiler e2)@[VM_AND]
           |compiler (OR(e1,e2)) = (compiler e1)@(compiler e2)@[VM_OR]
           |compiler (IMP(e1,e2)) = (compiler e1)@(compiler e2)@[VM_IMP]
           |compiler (EQI(e1,e2)) = (compiler e1)@(compiler e2)@[VM_EQI]
           |compiler (NOT e) = (compiler e)@[VM_NOT]
           |compiler (EQ(e1,e2)) = (compiler e1)@(compiler e2)@[VM_EQ]
           |compiler (LT(e1,e2)) = (compiler e1)@(compiler e2)@[VM_LT]
           |compiler (LTE(e1,e2)) = (compiler e1)@(compiler e2)@[VM_LTE]
           |compiler (GT(e1,e2)) = (compiler e1)@(compiler e2)@[VM_LTE,VM_NOT]
           |compiler (GTE(e1,e2)) = (compiler e1)@(compiler e2)@[VM_LT,VM_NOT]
           |compiler (ASMT(s,e)) = (symt := (insert s (!symt));(compiler e)@[(VAR s)])
           |compiler (STMT e) = (compiler e)@[END]
           |compiler (SEQ(e1,e2)) = (compiler e1)@(compiler e2)
        
        fun error tycon_mismatch n = (
                                        print ("Static type mismatch in exp "^(Int.toString n));
                                        raise compiletime_error
                                     )
           |error (undefined_variable x) n = (
                                                  print ("Variable "^x^" in exp "^(Int.toString n)^" was never declared.");
                                                  raise compiletime_error
                                             )
           |error ex n = raise ex

        fun helper (SEQ(e1,e2)) n = (let
                                        val _ = statictypeof e1 
                                in
                                        ((compiler e1)@(helper e2 (n+1))) 
                                end handle ex => (error ex n))
           |helper e n = (let
                                val _ = statictypeof e 
                         in
                           (compiler e)
                         end handle ex => (error ex n))
        in
        fun compile e = ((symt := ["c","e","Pi"]);(helper e 1)) (*handle ex =>   raise compiletime_error*)
        end;
        
        (* Stack operations : push, pop and top *)
        fun push m = (stack := m::(!stack));

        local
        fun stpop [] = raise stack_empty
           |stpop (x::xs) = ((stack := xs); x)
        in
        fun pop () = stpop (!stack)
        end;

        local
        fun sttop [] = raise stack_empty
           |sttop (x::xs) = x
        in
        fun top () = sttop (!stack)
        end;
       
        (* exec : The single instruction execution machinery. Stores result on
         * the top of the stack and unloads stack on encountering 'END'
         *)
        local
        (* ALU operations supported by MyVM *)
        fun add (MI a) (MI b) = MI(a+b)
           |add (MF a) (MF b) = MF(a+b)
           |add (MS a) (MS b) = MS(a^b)
           |add _ _ = raise tycon_mismatch

        fun sub (MI a) (MI b) = MI(a-b)
           |sub (MF a) (MF b) = MF(a-b)
           |sub _ _ = raise tycon_mismatch
        
        fun mult (MI a) (MI b) = MI(a*b)
           |mult (MF a) (MF b) = MF(a*b)
           |mult _ _ = raise tycon_mismatch

        fun divide (MI a) (MI b) = if(b <> 0) then MI(a div b) else raise math_error
           |divide (MF a) (MF b) = if(Real.==(b,0.0)) then raise math_error else MF(a/b)
           |divide _ _ = raise tycon_mismatch

        fun modulo (MI a) (MI b) = if(b <> 0) then MI(a mod b) else raise math_error
           |modulo _ _ = raise tycon_mismatch

        fun and_ (MB x) (MB y) = MB(x andalso y)
           |and_ _ _ = raise tycon_mismatch

        fun or_ (MB x) (MB y) = MB(x andalso y)
           |or_ _ _ = raise tycon_mismatch
        
        fun not_ (MB x) = MB(not x)
           |not_ _ = raise tycon_mismatch

        fun imp (MB x) (MB y) = MB((not x) orelse y)
           |imp _ _ = raise tycon_mismatch

        fun eqi (MB x) (MB y) = MB(x=y)
           |eqi _ _ = raise tycon_mismatch

        fun eq (MI a) (MI b) = MB(a=b)
           |eq (MF a) (MF b) = MB(Real.==(a,b))
           |eq _ _ = raise tycon_mismatch

        fun lt (MI a) (MI b) = MB(a<b)
           |lt (MF a) (MF b) = MB(a<b)
           |lt _ _ = raise tycon_mismatch

        fun lte (MI a) (MI b) = MB(a<=b)
           |lte (MF a) (MF b) = MB(a<=b)
           |lte _ _ = raise tycon_mismatch

        fun alu f = let 
                        val v1 = pop()
                        val v2 = pop()
                    in
                        (f v2 v1)
                    end
        in
        fun exec (LDI n) = push(MI n)
           |exec (LDR r) = push(MF r)
           |exec (LDB b) = push(MB b)
           |exec (LDS s) = push(MS s)
           |exec (LDV v) = push(lookup(v))
           |exec VM_ADD = push(alu add)
           |exec VM_SUB = push(alu sub)
           |exec VM_MUL = push(alu mult)
           |exec VM_DIV = push(alu divide)
           |exec VM_MOD = push(alu modulo)
           |exec VM_AND = push(alu and_)
           |exec VM_OR = push(alu or_)
           |exec VM_NOT = push(not_ (pop()))
           |exec VM_IMP = push(alu imp)
           |exec VM_EQI = push(alu eqi)
           |exec VM_EQ = push(alu eq)
           |exec VM_LT = push(alu lt)
           |exec VM_LTE = push(alu lte)
           |exec (VAR s) = push(assign s (pop()))
           |exec END = print (m2s (pop()))
        end;
 
        (* execute : The processor. 
         * Executes a sequence of instructions sequentially. 
         * Aborts execution on encountering any runtime error(s).
         * Runtime errors include dynamic type errors and undefined variables.
         * Initialises VM at every call.
         *)
        local
        fun helper [] n = print "EOF\nProgram exited normally. VM unloaded.\n"
           |helper (END::cs) n = (
                                        (exec END);
                                        print("\n["^(Int.toString n)^"]\t"); 
                                        (helper cs (n+1))
                                 )
           |helper (c::cs) n = ((exec c); (helper cs n))
        in
        fun execute code = ((init ());print "[0]\t";(helper code 1))
                handle (tycon_mismatch) => (print "Dynamic type mismatch.\nVM killed.";raise runtime_error)
                     | (undefined_variable x) => (print ("Variable "^x^" undefined.\nVM killed.");raise runtime_error)
                     | (math_error) => (print "Mathematical Error.\nVM killed.";raise runtime_error)
        end;

        val st = fn () => (!stack);
end;

